import pickle
import time
from typing import Any, List, Tuple, Union

import zmq

from easypubsub.logging import getLogger


class Subscriber:
    def __init__(
        self,
        name: str,
        proxy_subscribers_address: str,
        topics: Union[str, List[str]] = "",
        receive_timeout: float = 0.1,
    ) -> None:

        self.name = name
        self.subscribers_address = proxy_subscribers_address
        self.receive_timeout_ms = round(receive_timeout * 1000)

        if topics == "":
            self.topics = []
        elif isinstance(topics, str):
            self.topics = [topics.encode("utf-8")]
        else:
            self.topics = [topic.encode("utf-8") for topic in topics]

        self._logger = getLogger(f"EasyPubSub.Subscriber({name})")

        self.connect()

    def __del__(self) -> None:
        self.poller.unregister(self.socket)

    def connect(self) -> None:
        self.ctx = zmq.Context.instance()
        self.socket = self.ctx.socket(zmq.SUB)
        self._logger.info(f"Connecting to {self.subscribers_address}.")
        self.socket.connect(self.subscribers_address)

        if len(self.topics) > 0:
            for topic in self.topics:
                self._logger.info(f"Subscribing to {topic.decode('utf-8')}.")
                self.socket.setsockopt(zmq.SUBSCRIBE, topic)
        else:
            self._logger.info("Subscribing to all topics.")
            self.socket.setsockopt(zmq.SUBSCRIBE, b"")

        self.poller = zmq.Poller()
        self.poller.register(self.socket, zmq.POLLIN)

        time.sleep(1)

    def receive(self) -> List[Tuple[str, Any]]:
        messages: List[Any] = []
        messages_available = True
        while messages_available:
            sockets = dict(self.poller.poll(self.receive_timeout_ms))
            if self.socket in sockets:
                topic, message = self.socket.recv_multipart()
                messages.append((topic.decode("utf-8"), pickle.loads(message)))
            else:
                messages_available = False

        return messages



import time
from easypubsub.publisher import Publisher
from easypubsub.subscriber import Subscriber
from easypubsub.proxy import Proxy

def test_simple_pubsub():
    """
    Test the simple Publish/Subscribe functionality.
    """

    SUBSCRIBERS_ADDRESS = "tcp://127.0.0.1:5555"
    PUBLISHERS_ADDRESS = "tcp://127.0.0.1:5556"

    # Create a Proxy.
    proxy = Proxy(SUBSCRIBERS_ADDRESS, PUBLISHERS_ADDRESS)
    proxy.launch()
    time.sleep(1.0)

    # Create a Publisher.
    publisher = Publisher("test_publisher", PUBLISHERS_ADDRESS, default_topic="test_topic")
    publisher.publish("This is a first test message.")

    # Create a Subscriber.
    subscriber = Subscriber("test_subscriber", SUBSCRIBERS_ADDRESS, topics="test_publisher.test_topic")

    # Wait for connection to establish.
    time.sleep(1.0)
    publisher.publish("This is a second test message.")
    messages = subscriber.receive()
    assert len(messages) == 1

    # Stop the Proxy.
    proxy.stop()


import pickle
import time
from typing import Any, Optional

import zmq

from easypubsub.logging import getLogger


class Publisher:
    def __init__(
        self, name: str, proxy_publishers_address: str, default_topic: str = ""
    ) -> None:
        self.publishers_address = proxy_publishers_address
        self.default_topic = default_topic
        self.name = name

        self._logger = getLogger(f"EasyPubSub.Publisher({name})")
        self.connect()

    def connect(self) -> None:
        self.ctx = zmq.Context.instance()
        self.socket = self.ctx.socket(zmq.PUB)
        self._logger.info(f"Connecting to {self.publishers_address}.")
        self.socket.connect(self.publishers_address)

        time.sleep(1)

    def publish(self, message: Any, topic: Optional[str] = None) -> None:
        if topic is None:
            topic = self.default_topic
        if topic.endswith("."):
            self._logger.warning(
                f'Topic "{topic}" ends with a dot, I will remove the final dot before publishing.'
            )
        topic = f"{self.name}.{topic}".strip(".")
        try:
            pickled_message = pickle.dumps(message)
            self.socket.send_multipart([topic.encode("utf-8"), pickled_message])
        except:
            self._logger.exception("Could not publish message. See traceback.")



import time
from easypubsub.proxy import Proxy

def test_proxy():
    SUBSCRIBERS_ADDRESS = "tcp://127.0.0.1:5555"
    PUBLISHERS_ADDRESS = "tcp://127.0.0.1:5556"
    proxy = Proxy(SUBSCRIBERS_ADDRESS, PUBLISHERS_ADDRESS)
    proxy.launch()
    time.sleep(0.5)
    proxy.stop()



from threading import Thread
from typing import Optional

import zmq
from zmq.devices import ThreadProxy
from zmq.utils.win32 import allow_interrupt

from easypubsub.logging import getLogger

_logger = getLogger("EasyPubSub.Proxy")


class Proxy:
    """The EasyPubSub Proxy acts as an intermediary between Publishers and Subscribers.

    Args:
        subscribers_address: The address that subscribers will use to connect to the Proxy.
        publishers_address: The address that publisher will use to connect to the Proxy.

    Example:
        >>> from easypubsub.proxy import Proxy
        >>> proxy = Proxy("tcp://localhost:5555", "tcp://localhost:5556")
    """

    def __init__(self, subscribers_address: str, publishers_address: str) -> None:
        self.ctx = zmq.Context.instance()
        self.subcriber_address = subscribers_address
        self.publishers_address = publishers_address

        _logger.info("Creating socket for subscribers.")
        self.xpub_subscriber_socket = self.ctx.socket(zmq.XPUB)
        _logger.info(
            "Binding socket for subscribers to {}.".format(self.subcriber_address)
        )
        self.xpub_subscriber_socket.bind(self.subcriber_address)

        _logger.info("Creating socket for publishers.")
        self.xsub_publisher_socket = self.ctx.socket(zmq.XSUB)
        _logger.info(
            "Binding socket for publishers to {}.".format(self.publishers_address)
        )
        self.xsub_publisher_socket.bind(self.publishers_address)

        self._proxy_thread: Optional[Thread] = None

    def stop(self) -> None:
        """Fix CTRL-C on Windows."""

        # self.ctx.term()
        self.xpub_subscriber_socket.close()
        self.xsub_publisher_socket.close()

        if self._proxy_thread is not None and self._proxy_thread.is_alive():
            self._proxy_thread.join()

    def _launch(self) -> None:
        _logger.info("Launching proxy.")
        try:
            with allow_interrupt(self.stop):
                zmq.proxy(self.xpub_subscriber_socket, self.xsub_publisher_socket)
        except KeyboardInterrupt:
            _logger.info("Detected KeyboardInterrupt. Closing proxy and sockets.")
            self.xpub_subscriber_socket.close()
            self.xsub_publisher_socket.close()
        except zmq.error.ContextTerminated:
            _logger.info("Detected ContextTerminated. Closing proxy and sockets.")
            self.xpub_subscriber_socket.close()
            self.xsub_publisher_socket.close()
        except:
            _logger.exception("Unhandled exception. Closing proxy and sockets.")
            self.xpub_subscriber_socket.close()
            self.xsub_publisher_socket.close()

        _logger.info("Done.")

    def launch(self) -> None:
        """Launch the Proxy."""

        if self._proxy_thread is not None:
            _logger.warning("Proxy already launched.")
        else:
            self._proxy_thread = Thread(target=self._launch)
            self._proxy_thread.start()
